from dataclasses import dataclass
import imp
import json
import asyncio
import inspect
import traceback
import httpx
import aiocache
from typing import Any, Dict, List, Type, Union, Callable, Optional, cast

from pygtrie import StringTrie
from nonebot.typing import overrides
from nonebot.log import logger
from nonebot.exception import WebSocketClosed
from nonebot.utils import DataclassEncoder, escape_tag
from nonebot.drivers import (
    URL,
    Driver,
    Request,
    Response,
    WebSocket,
    ForwardDriver,
    ReverseDriver,
    HTTPServerSetup,
    WebSocketServerSetup,
)

from nonebot.adapters import Adapter as BaseAdapter

from . import event
from .bot import Bot
from .config import Config as TelegramConfig
from .event import (
    Event,
    GroupMessageEvent,
    MessageEvent,
    NewChatMembersEvent,
    PrivateMessageEvent,
    LeafChatMemberEvent,
    NewChatTitleEvent,
    DeleteChatPhotoEvent,
    VoiceChatEndedEvent,
    VoiceChatStartedEvent,
    CallbackQueryEvent
)
from .message import Message, MessageSegment
from .exception import NetworkError, ApiNotAvailable, ActionFailed, TelegramAdapterConfigException, MessageNotAcceptable
from .utils import log
from .cache import TelegramCache

from nonebot.message import handle_event


class Adapter(BaseAdapter):

    telegram_config: TelegramConfig
    bot_name: str
    use_long_polling: bool
    cache: TelegramCache


    @overrides(BaseAdapter)
    def __init__(self, driver: Driver, **kwargs: Any):
        super().__init__(driver, **kwargs)
        self.telegram_config: TelegramConfig = TelegramConfig(
            **self.config.dict())
        self.tasks: List["asyncio.Task"] = []
        self._check_config()
        #loop = asyncio.get_event_loop()
        #loop.run_until_complete(self._call_api(None, "deleteWebhook"))
        #self.bot_name = loop.run_until_complete(self._call_api(None, "getMe"))["username"]
        # setup route
        log("INFO", f"Current driver:{self.driver.type}")
        if "httpx" in self.driver.type:
            self.driver.on_startup(self._start_polling)
            self.driver.on_shutdown(self._stop_polling)
        elif "fastapi" in self.driver.type:
            http_setup = HTTPServerSetup(
                URL(f"/{self.telegram_config.bot_token}"), "POST", self.get_name(), self._handle_webhook
            )
            self.setup_http_server(http_setup)
            self.driver.on_startup(self._setup_webhook)
        else:
            raise TelegramAdapterConfigException("No avaliable driver type")
        #setup cache
        self.cache = TelegramCache()

    @classmethod
    @overrides(BaseAdapter)
    def get_name(cls) -> str:
        """适配器名称: `Telegram`"""
        return "Telegram"

    @overrides(BaseAdapter)
    async def _call_api(self,
                        bot: Bot,
                        api: str,
                        **data) -> Any:
        # if self.connection_type != "http":
        #    log("ERROR", "Only support http connection.")
        #    return
        # 将方法名称改为驼峰式 from nonebot/adapter-telegram
        api = api.split("_", maxsplit=1)[0] + "".join(
            s.capitalize() for s in api.split("_")[1:]
        )
        if self.use_long_polling and api == "getUpdates":
            api_timeout = self.telegram_config.telegram_long_polling_timeout
        else:
            api_timeout=self.config.api_timeout
        log("DEBUG", f"Calling API <y>{api}</y>")
        # print(data)
        headers = {"Content-Type": "application/json"}
        if not data:
            data = {}
            #raise ValueError("data not found")
        if data.get("data") != None and len(data) == 1:
            data = data.get("data")
        try:
            async with httpx.AsyncClient(headers=headers,proxies=self.telegram_config.telegram_bot_api_proxy) as client:
                response = await client.post(f"{self.telegram_config.telegram_bot_api_server_addr}/bot{self.telegram_config.bot_token}/{api}",
                                             json=data,
                                             timeout=api_timeout)
            if 200 <= response.status_code < 500:
                result = response.json()
                if isinstance(result, dict):
                    if result.get("ok") != True:
                        #print(result)
                        raise ActionFailed(result.get("error_code"),result.get("description"))
                    # print(result["result"])
                    return result["result"]
            raise NetworkError(f"HTTP request received unexpected "
                               f"status code: {response.status_code}")
        except httpx.InvalidURL:
            raise NetworkError("API root url invalid")
        except httpx.HTTPError:
            raise NetworkError("HTTP request failed")

    async def _call_multipart_form_data_api(self, api: str, file: dict, data: dict):
        log("DEBUG", f"Calling API <y>{api}</y>")
        #print(data)
        for key in data:
            if isinstance(data[key], int) or isinstance(data[key], float):
                data[key] = str(data[key])
            elif not isinstance(data[key], str):
                try:
                    data[key] = json.dumps(data[key])
                except:
                    del data[key]
        try:
            async with httpx.AsyncClient(proxies=self.telegram_config.telegram_bot_api_proxy) as client:
                response = await client.post(f"{self.telegram_config.telegram_bot_api_server_addr}/bot{self.telegram_config.bot_token}/{api}",
                                             files=file,
                                             data=data,
                                             timeout=self.config.api_timeout)
            if 200 <= response.status_code < 500:
                result = response.json()
                #print(result)
                if isinstance(result, dict):
                    if result.get("ok") != True:
                        raise ActionFailed()
                    #print(result["result"])
                    return result["result"]
            #print(result)
            raise NetworkError(f"HTTP request received unexpected "
                               f"status code: {response.status_code}")
        except httpx.InvalidURL:
            raise NetworkError("API root url invalid")
        except httpx.HTTPError:
            raise NetworkError("HTTP request failed")
        finally:
                for file_obj in file.values():
                    try:
                        file_obj.close()
                    except:
                        pass

    def _check_config(self):
        if not self.telegram_config.bot_token:
            raise TelegramAdapterConfigException("bot token not set")
        if self.telegram_config.telegram_polling_interval == 0 and self.telegram_config.telegram_long_polling_timeout == 0:
            raise TelegramAdapterConfigException(
                "telegram_polling_interval and telegram_long_polling_timeout are both 0, please check your config")
        if self.telegram_config.telegram_polling_interval != 0 and self.telegram_config.telegram_long_polling_timeout != 0:
            raise TelegramAdapterConfigException(
                "telegram_polling_interval and telegram_long_polling_timeout are both not 0, please check your config")
        if self.telegram_config.telegram_polling_interval == 0:
            self.use_long_polling = True
        else:
            self.use_long_polling = False

    async def _setup_webhook(self):
        await self._call_api(None, "deleteWebhook")
        username = (await self._call_api(None, "getMe"))["username"]
        self.bot_name = username
        await self._call_api(None, "setWebhook", url=f"{self.telegram_config.webhook_addr}/{self.telegram_config.bot_token}/")

    async def _handle_webhook(self, request: Request) -> Response:
        data = request.content
        json_data = json.loads(data)
        event = self.json_to_event(json_data)
        try:
            await handle_event(Bot(self, self.bot_name), event)
        except Exception as e:
            logger.opt(colors=True, exception=e).error(
                f"<r><bg #f8bbd0>Failed to handle event. Raw: {json_data}</bg #f8bbd0></r>"
            )
        return Response(200)

    def _pre_process_event(self, event: MessageEvent):
        if isinstance(event, GroupMessageEvent):
            if event.message.text:
                if f"@{self.bot_name}" in event.message.text:
                    event.message.text = event.message.text.replace(
                        f"@{self.bot_name}", "").strip()
                    event.to_me = True
                    print(f"event.message.text:{event.message.text}")

    def json_to_event(self, json_data: Any) -> Optional[Event]:
        try:
            #print(json_data)
            if "callback_query" in json_data:
                if json_data["callback_query"]["from"]["is_bot"]:
                    return
                event = CallbackQueryEvent.parse_obj(json_data)
            elif "message" in json_data:
                if json_data["message"]["from"]["is_bot"]:
                    return
                json_data["user_id"] = json_data["message"]["from"]["id"]
                json_data["group_id"] = json_data["message"]["chat"]["id"]
                if json_data["message"]["chat"]["type"] == "private":
                    event = PrivateMessageEvent.parse_obj(json_data)
                elif "group" in json_data["message"]["chat"]["type"]:
                    if "new_chat_members" in json_data["message"]:
                        event = NewChatMembersEvent.parse_obj(json_data)
                    elif "left_chat_member" in json_data["message"]:
                        event = LeafChatMemberEvent.parse_obj(json_data)
                    elif "new_chat_title" in json_data["message"]:
                        event = NewChatTitleEvent.parse_obj(json_data)
                    elif "voice_chat_started" in json_data["message"]:
                        event = VoiceChatStartedEvent.parse_obj(json_data)
                    elif "voice_chat_ended" in json_data["message"]:
                        event = VoiceChatEndedEvent.parse_obj(json_data)
                    else:
                        event = GroupMessageEvent.parse_obj(json_data)
            else:
                raise MessageNotAcceptable("")
                #event = MessageEvent.parse_obj(message)
            self._pre_process_event(event)
            return event
        except Exception as e:
            log("ERROR", "Event Parser Error", e)
            raise MessageNotAcceptable("")

    async def _stop_polling(self) -> None:
        try:
            for task in self.tasks:
                if not task.done():
                    task.cancel()

            await asyncio.gather(*self.tasks, return_exceptions=True)
        except:
            pass

    async def _start_polling(self):
        async def polling_handle_event(message, bot, event):
            try:
                await handle_event(bot, event)
            except Exception as e:
                logger.opt(colors=True, exception=e).error(
                    f"<r><bg #f8bbd0>Failed to handle event. Raw: {message}</bg #f8bbd0></r>"
                )

        async def polling():
            await self._call_api(None, "deleteWebhook")
            username = (await self._call_api(None, "getMe"))["username"]
            self.bot_name = username
            bot = Bot(self, self.bot_name)
            log("INFO", "Reset Update...")
            await self._call_api(None, "getUpdates", offset=-1, timeout=self.telegram_config.telegram_long_polling_timeout)
            offset: int = 0
            log("INFO", "Start polling")
            while True:
                try:
                    messages = await self._call_api(None, "getUpdates", offset=offset if not offset == 0 else None, timeout=self.telegram_config.telegram_long_polling_timeout)
                    for message in messages:
                        if offset < message["update_id"] + 1:
                            offset = message["update_id"] + 1
                            #print(f"offset update to {offset}")
                        event = self.json_to_event(message)
                        try:
                            loop = asyncio.get_event_loop() 
                            loop.create_task(polling_handle_event(message, bot, event))
                        except Exception as e:
                            logger.opt(colors=True, exception=e).error(
                                f"<r><bg #f8bbd0>Failed to handle event. Raw: {message}</bg #f8bbd0></r>"
                            )
                except (KeyboardInterrupt,asyncio.exceptions.CancelledError):
                    break
                except NetworkError:
                    if not self.use_long_polling:
                        log("ERROR", "Failed to handle polling")
                        traceback.print_exc()
                except:
                        log("ERROR", "Failed to handle polling")
                        traceback.print_exc()
                if not self.use_long_polling:
                    await asyncio.sleep(self.telegram_config.telegram_polling_interval)

        log("INFO", "Setting up polling...")
        self.tasks.append(asyncio.create_task(polling()))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot', 'nonebot.adapters.telegram']

package_data = \
{'': ['*']}

install_requires = \
['aiocache>=0.11.1,<0.12.0', 'httpx==0.20.0', 'nonebot2>=2.0.0-beta.2,<3.0.0']

setup_kwargs = {
    'name': 'nonebot-adapter-antelegram',
    'version': '0.2.0.dev4',
    'description': 'Another unofficial Telegram adapter for nonebot2',
    'long_description': '# nonebot-adapter-telegram\n（施工中）自己用的非官方nonebot2 telegram adapter，代码全靠糊  \n开发中代码没有经过清理和优化，不能与官方版本共存  \n当前仅支持有限类型的消息解析和发送（接受私聊/群聊文字/图片，发送私聊/群聊文字/图片/语音，入群事件）  \n如果使用webhook工作方式需要公网ip或者frp  \n演示bot[@aya_od_bot](https://t.me/aya_od_bot)  \n## 使用方法\n如果要试毒的话  \n真的要的话  \n```shell\npip install nonebot-adapter-antelegram\n```\n## 上路\n一、  \n新建项目文件夹  \n二、  \n在nonebot2的配置文件中配置以下选项  \n```shell\nbot_token=your_bot_token  #telegram bot token，需要事先申请，参考https://core.telegram.org/bots#3-how-do-i-create-a-bot\ntelegram_bot_api_server_addr=https://api.telegram.org #可选，应该大概也可以替换为反代的域名，不设置默认官方\ntelegram_bot_api_proxy=proxy_server_addr #可选，代理服务器地址\n\n#如果需要使用webhook方式接受消息，进行如下设置（推荐但是麻烦）\ndriver=~fastapi\nhost=127.0.0.1 # 配置 NoneBot 监听的 IP / 主机名  \nport=xxxxx     # 配置 NoneBot 监听的端口  \nwebhook_host=https://your_domain # 配置telegram webhook域名，由于telegram要求webhook地址必须为https，需要自行配置反向代理，也可以参考telegram文档自建本地bot api，本地api无需https  \n\n#如果需要长轮训方式接受消息，进行如下设置（推荐，长轮询比轮训接受消息更及时，资源占用更小）\ndriver=~httpx\ntelegram_polling_interval=0 #不使用轮训\ntelegram_long_polling_timeout=20 #长轮训超时时间\n\n#如果需要轮训方式接受消息，进行如下设置（不推荐，仅建议调试网络时使用）\ndriver=~httpx\ntelegram_polling_interval=5 #轮训间隔\ntelegram_long_polling_timeout=0 #不使用长轮训\n\n#注意：使用driver=~fastapi+~httpx时会使用httpx作为Driver并启动fastapi，用于需要轮训且同时启动fastapi服务器的情况\n\n```\n三、（仅使用webhook方式需要）  \n将webhook域名解析到本机，用你喜欢的方式配置反代将webhook域名的流量转发到nonebot2的监听端口（如果不使用本地bot api）  \n四、  \n开始写机器人（摸鱼）  \n\n## 已知问题（短时间内并不会解决）  \n仅支持接受有限种类的消息  \n仅支持发送有限种类的消息  \n有亿点点小bug  \n~~可能存在内存泄漏问题~~  \n\n## 最简单的例子\nbot.py\n```python\nimport nonebot\nfrom nonebot.adapters.telegram.adapter import Adapter\n\nnonebot.init()\ndriver = nonebot.get_driver()\ndriver.register_adapter(Adapter)\nnonebot.load_plugin("plugins.echo")\n\nif __name__ == "__main__":\n    nonebot.run()   \n```\nplugins/echo.py\n```python\nfrom nonebot.plugin import on_command\nfrom nonebot.adapters.telegram import Bot, MessageEvent, Message, MessageSegment\nfrom nonebot.rule import to_me\n\necho = on_command("echo",to_me())\n\n@echo.handle()\nasync def echo_escape(bot: Bot, event: MessageEvent):\n    await bot.send(event, event.get_message())\n\n#await bot.send(event, "114514") #发送文字\n#await bot.send(event, MessageSegment.photo(pic_url)) #发送图片 支持file:///，base64://，bytes，file_id，url(由Telegram服务器下载)  \n#\n```\n运行机器人，向bot私聊发送/echo 123，bot会将消息原样重新发送  \n\n\n',
    'author': 'ColdThunder11',
    'author_email': 'lslyj27761@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ColdThunder11/nonebot-adapter-telegram',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)
